"""Functions for creating tokens."""

import base64
import os

__all__ = ["create_token"]


def create_token() -> str:
    """Create a new random Gafaelfawr token.

    Normally, users of Gafaelfawr should use the Gafaelfawr API to create new
    tokens. This function is intended only for creating new bootstrap tokens
    that will be injected into the Gafaelfawr server via configuration, or for
    creating syntactically valid tokens for use with the Gafaelfawr mock.

    Returns
    -------
    str
        New random Gafaelfawr token. This token will not be registered with
        any running Gafaelfawr instance and therefore will not be usable
        without other measures, such as configuring it as a bootstrap token.
    """
    key = base64.urlsafe_b64encode(os.urandom(16)).decode().rstrip("=")
    secret = base64.urlsafe_b64encode(os.urandom(16)).decode().rstrip("=")
    return f"gt-{key}.{secret}"



"""Test fixtures for Gafaelfawr client testing."""

from pathlib import Path

import pytest
import pytest_asyncio
import respx
from rubin.repertoire import Discovery, register_mock_discovery

from rubin.gafaelfawr import MockGafaelfawr, register_mock_gafaelfawr


@pytest.fixture
def mock_discovery(
    respx_mock: respx.Router, monkeypatch: pytest.MonkeyPatch
) -> Discovery:
    monkeypatch.setenv("REPERTOIRE_BASE_URL", "https://example.com/repertoire")
    path = Path(__file__).parent / "data" / "discovery.json"
    return register_mock_discovery(respx_mock, path)


@pytest_asyncio.fixture
async def mock_gafaelfawr(
    mock_discovery: Discovery, respx_mock: respx.Router
) -> MockGafaelfawr:
    return await register_mock_gafaelfawr(respx_mock)



"""Tests for the Gafaelfawr mock.

Most of the functionality of the mock is tested by the regular client tests,
but test a few additional behaviors that the client tests didn't need or that
are specific to the mock.
"""

from datetime import UTC, datetime, timedelta
from pathlib import Path

import pytest
import respx
from rubin.repertoire import register_mock_discovery

from rubin.gafaelfawr import (
    GafaelfawrClient,
    GafaelfawrUserInfo,
    GafaelfawrWebError,
    MockGafaelfawr,
    MockGafaelfawrAction,
    register_mock_gafaelfawr,
)

from .support.data import read_test_user_info


@pytest.mark.asyncio
async def test_bad_token(mock_gafaelfawr: MockGafaelfawr) -> None:
    client = GafaelfawrClient()
    mock_gafaelfawr.set_user_info("someuser", read_test_user_info("someuser"))

    # Making a request with an invalid token should return a 403 error.
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.get_user_info("some-token", "someuser")
    assert exc_info.value.status == 403


@pytest.mark.asyncio
async def test_create_token_expires(mock_gafaelfawr: MockGafaelfawr) -> None:
    client = GafaelfawrClient()
    mock_gafaelfawr.set_user_info("someuser", read_test_user_info("someuser"))

    # Creating a token that is already expired should be reflected in the
    # token information and cause the token to be rejected in later calls.
    token = mock_gafaelfawr.create_token(
        "someuser", expires=datetime.now(tz=UTC) - timedelta(seconds=2)
    )
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.get_user_info(token)
    assert exc_info.value.status == 401


@pytest.mark.asyncio
async def test_fail_on(mock_gafaelfawr: MockGafaelfawr) -> None:
    client = GafaelfawrClient()
    token = mock_gafaelfawr.create_token("someuser")
    mock_gafaelfawr.set_user_info("someuser", read_test_user_info("someuser"))
    admin_token = mock_gafaelfawr.create_token(
        "admin", scopes=["admin:userinfo", "admin:token"]
    )
    admin_userinfo = GafaelfawrUserInfo(username="admin")
    mock_gafaelfawr.set_user_info("admin", admin_userinfo)

    # Setting failure for user info for a user should block both ways of
    # retrieving the user information.
    mock_gafaelfawr.fail_on("someuser", MockGafaelfawrAction.USER_INFO)
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.get_user_info(token)
    assert exc_info.value.status == 500
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.get_user_info(admin_token, "someuser")
    assert exc_info.value.status == 500

    # Getting the user information for the admin token works.
    assert await client.get_user_info(admin_token) == admin_userinfo

    # Check failure for creating a token.
    mock_gafaelfawr.fail_on("admin", MockGafaelfawrAction.CREATE_TOKEN)
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.create_service_token(admin_token, "bot-user", scopes=[])


@pytest.mark.asyncio
async def test_missing_discovery(
    monkeypatch: pytest.MonkeyPatch, respx_mock: respx.Router
) -> None:
    monkeypatch.setenv("REPERTOIRE_BASE_URL", "https://example.com/repertoire")
    path = Path(__file__).parent / "data" / "empty.json"
    register_mock_discovery(respx_mock, path)

    # If the Gafaelfawr service (v1) is not included in the discovery
    # information, register_mock_gafaelfawr should assert.
    with pytest.raises(AssertionError):
        await register_mock_gafaelfawr(respx_mock)


@pytest.mark.asyncio
async def test_required_scopes(mock_gafaelfawr: MockGafaelfawr) -> None:
    client = GafaelfawrClient()
    mock_gafaelfawr.set_user_info("someuser", read_test_user_info("someuser"))

    # A token without admin:userinfo should not be able to access the user
    # route used by get_user_info with an explicit username.
    token = mock_gafaelfawr.create_token("otheruser")
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.get_user_info(token, "someuser")
    assert exc_info.value.status == 403

    # A token without admin:token should not be able to create a token.
    token = mock_gafaelfawr.create_token("admin", scopes=["admin:userinfo"])
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.create_service_token(token, "bot-user", scopes=[])
    assert exc_info.value.status == 403


@pytest.mark.asyncio
async def test_service_token_expiration(
    mock_gafaelfawr: MockGafaelfawr,
) -> None:
    token = mock_gafaelfawr.create_token("admin", scopes=["admin:token"])
    client = GafaelfawrClient()

    # The expiration of a service token should be checked in later mock calls,
    # such as getting user information for the token.
    service_token = await client.create_service_token(
        token,
        "bot-user",
        expires=datetime.now(tz=UTC) - timedelta(seconds=2),
        scopes=[],
    )
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.get_user_info(service_token)
    assert exc_info.value.status == 401


@pytest.mark.asyncio
async def test_service_token_username(mock_gafaelfawr: MockGafaelfawr) -> None:
    token = mock_gafaelfawr.create_token("admin", scopes=["admin:token"])
    client = GafaelfawrClient()

    # Service token usernames must begin with "bot-".
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.create_service_token(token, "user", scopes=[])
    assert exc_info.value.status == 422
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.create_service_token(token, "botuser", scopes=[])
    assert exc_info.value.status == 422



"""Client for Gafaelfawr."""

from ._client import GafaelfawrClient
from ._dependencies import GafaelfawrDependency, gafaelfawr_dependency
from ._exceptions import (
    GafaelfawrDiscoveryError,
    GafaelfawrError,
    GafaelfawrNotFoundError,
    GafaelfawrValidationError,
    GafaelfawrWebError,
)
from ._mock import (
    MockGafaelfawr,
    MockGafaelfawrAction,
    register_mock_gafaelfawr,
)
from ._models import (
    GafaelfawrGroup,
    GafaelfawrNotebookQuota,
    GafaelfawrQuota,
    GafaelfawrTapQuota,
    GafaelfawrUserInfo,
)
from ._tokens import create_token

__all__ = [
    "GafaelfawrClient",
    "GafaelfawrDependency",
    "GafaelfawrDiscoveryError",
    "GafaelfawrError",
    "GafaelfawrGroup",
    "GafaelfawrNotFoundError",
    "GafaelfawrNotebookQuota",
    "GafaelfawrQuota",
    "GafaelfawrTapQuota",
    "GafaelfawrUserInfo",
    "GafaelfawrValidationError",
    "GafaelfawrWebError",
    "MockGafaelfawr",
    "MockGafaelfawrAction",
    "create_token",
    "gafaelfawr_dependency",
    "register_mock_gafaelfawr",
]



"""FastAPI dependencies providing a Gafaelfawr client."""

from typing import Annotated

from fastapi import Depends
from httpx import AsyncClient
from rubin.repertoire import DiscoveryClient, discovery_dependency
from safir.dependencies.http_client import http_client_dependency

from ._client import GafaelfawrClient

__all__ = ["GafaelfawrDependency", "gafaelfawr_dependency"]


class GafaelfawrDependency:
    """Maintain a global Gafaelfawr client.

    This is structured as a dependency that creates and caches the client on
    first use to delay client creation until runtime so that the test suite
    has a chance to initialize environment variables.
    """

    def __init__(self) -> None:
        self._http_client: AsyncClient | None = None
        self._client: GafaelfawrClient | None = None

    async def __call__(
        self,
        discovery: Annotated[DiscoveryClient, Depends(discovery_dependency)],
        http_client: Annotated[AsyncClient, Depends(http_client_dependency)],
    ) -> GafaelfawrClient:
        if not self._client or self._http_client != http_client:
            self._client = GafaelfawrClient(
                http_client, discovery_client=discovery
            )
            self._http_client = http_client
        return self._client


gafaelfawr_dependency = GafaelfawrDependency()
"""The cached Gafaelfawr client as a FastAPI dependency."""



"""Mock for the parts of the Gafaelfawr API used by the client."""

# Required for now because Ruff removes explicitly quoted types and mypy
# otherwise gets confused about the type of the internal check decorator.
from __future__ import annotations

import re
from collections import defaultdict
from collections.abc import Callable, Iterable
from datetime import UTC, datetime
from enum import Enum
from functools import wraps
from typing import TYPE_CHECKING, Concatenate
from urllib.parse import urljoin

from httpx import Request, Response
from rubin.repertoire import DiscoveryClient

from ._models import AdminTokenRequest, GafaelfawrUserInfo, NewToken, TokenData
from ._tokens import create_token

if TYPE_CHECKING:
    import respx

type _GafaelfawrMockPublic[**P] = Callable[
    Concatenate[MockGafaelfawr, Request, P], Response
]
"""External type exposed as a side effect method."""

type _GafaelfawrMockInternal[**P] = Callable[
    Concatenate[MockGafaelfawr, Request, TokenData, P], Response
]
"""Internal type for the implementation of a side effect method."""

__all__ = [
    "MockGafaelfawr",
    "MockGafaelfawrAction",
    "register_mock_gafaelfawr",
]


class MockGafaelfawrAction(Enum):
    """Possible actions that could fail."""

    CREATE_TOKEN = "create_token"
    USER_INFO = "user_info"


class MockGafaelfawr:
    """Mock for the parts of the Gafaelfawr API used by the client."""

    def __init__(self) -> None:
        self._fail: defaultdict[str, set[MockGafaelfawrAction]]
        self._fail = defaultdict(set)
        self._tokens: dict[str, TokenData] = {}
        self._user_info: dict[str, GafaelfawrUserInfo | None] = {}

    def create_token(
        self,
        username: str,
        *,
        expires: datetime | None = None,
        scopes: Iterable[str] | None = None,
    ) -> str:
        """Create a token for the given username.

        This token will only be recognized by the same instance of the
        Gafaelfawr mock.

        Parameters
        ----------
        username
            Username the token is for.
        expires
            If provided, sets the expiration time of the token.
        scopes
            If provided, list of scopes to assign to the token. This is
            primarily needed if the client will call a privileged API
            endpoint.

        Returns
        -------
        str
            New Gafaelfawr token.
        """
        token = create_token()
        self._tokens[token] = TokenData(
            token=token,
            username=username,
            expires=expires,
            scopes=set(scopes) if scopes else set(),
        )
        return token

    def fail_on(
        self,
        username: str,
        actions: MockGafaelfawrAction | Iterable[MockGafaelfawrAction],
    ) -> None:
        """Configure the API to fail on requests for the given user.

        This can be used by test suites to test handling of Gafaelfawr
        failures.

        Parameters
        ----------
        username
            Username for which operations should fail.
        actions
            An action or iterable of actions that should fail. Pass in the
            empty list to restore regular operations for this user.
        """
        if isinstance(actions, MockGafaelfawrAction):
            self._fail[username] = {actions}
        else:
            self._fail[username] = set(actions)

    def install_routes(self, respx_mock: respx.Router, base_url: str) -> None:
        """Install the mock routes for the Gafaelfawr API.

        Parameters
        ----------
        respx_mock
            Mock router to use to install routes.
        base_url
            Base URL for the mock routes.
        """
        prefix = base_url.rstrip("/") + "/"
        handler = self._handle_user_info
        respx_mock.get(urljoin(prefix, "user-info")).mock(side_effect=handler)
        handler = self._handle_create_token
        respx_mock.post(urljoin(prefix, "tokens")).mock(side_effect=handler)

        # These routes require regex matching of the username.
        base_regex = re.escape(base_url.rstrip("/"))
        regex = re.compile(base_regex + "/users/(?P<username>[^/]+)$")
        respx_mock.get(url__regex=regex).mock(side_effect=self._handle_user)

    def set_user_info(
        self, username: str, user_info: GafaelfawrUserInfo | None
    ) -> None:
        """Set the user information for a given user.

        Parameters
        ----------
        username
            Username for which to set a quota.
        user_info
            User information to return for that user, or `None` to return a
            404 error.
        """
        assert user_info is None or user_info.username == username, (
            f"User info for wrong user ({user_info.username} != {username})"
        )
        self._user_info[username] = user_info

    @staticmethod
    def _check[**P](
        *,
        fail_on: MockGafaelfawrAction | None = None,
        required_scope: str | None = None,
    ) -> Callable[[_GafaelfawrMockInternal[P]], _GafaelfawrMockPublic[P]]:
        """Wrap `MockGafaelfawr` methods to perform common checks.

        There are various common checks that should be performed for every
        request to the mock, and the token always has to be extracted from the
        requst and injected as an additional argument to the method. This
        wrapper performs those checks and then injects the token data into the
        underlying handler.

        Paramaters
        ----------
        fail_on
            If this user is configured to fail on this action, return a
            failure rather than calling the underlying handler.

        Returns
        -------
        typing.Callable
            Decorator to wrap `MockGafaelfawr` methods.
        """

        def decorator(
            f: _GafaelfawrMockInternal[P],
        ) -> _GafaelfawrMockPublic[P]:
            @wraps(f)
            def wrapper(
                mock: MockGafaelfawr,
                request: Request,
                *args: P.args,
                **kwargs: P.kwargs,
            ) -> Response:
                authorization = request.headers["Authorization"]
                scheme, token = authorization.split(None, 1)
                if scheme.lower() != "bearer":
                    return Response(403)
                token_data = mock._tokens.get(token)
                if not token_data:
                    return Response(403)
                if fail_on and fail_on in mock._fail[token_data.username]:
                    return Response(500)
                if required_scope and required_scope not in token_data.scopes:
                    return Response(403)
                if token_data.expires:
                    if token_data.expires <= datetime.now(tz=UTC):
                        return Response(401)
                return f(mock, request, token_data, *args, **kwargs)

            return wrapper

        return decorator

    @_check(
        fail_on=MockGafaelfawrAction.CREATE_TOKEN, required_scope="admin:token"
    )
    def _handle_create_token(
        self, request: Request, token_data: TokenData
    ) -> Response:
        body = AdminTokenRequest.model_validate_json(request.content.decode())
        if not body.username.startswith("bot-"):
            return Response(422)
        token = self.create_token(
            body.username,
            expires=body.expires,
            scopes=body.scopes,
        )
        userinfo = GafaelfawrUserInfo(
            username=body.username,
            name=body.name,
            uid=body.uid,
            gid=body.gid,
            groups=body.groups or [],
        )
        self.set_user_info(body.username, userinfo)
        response = NewToken(token=token)
        return Response(200, json=response.model_dump(mode="json"))

    @_check(required_scope="admin:userinfo")
    def _handle_user(
        self,
        request: Request,
        token_data: TokenData,
        *,
        username: str,
    ) -> Response:
        if MockGafaelfawrAction.USER_INFO in self._fail[username]:
            return Response(500)
        elif user_info := self._user_info.get(username):
            result = user_info.model_dump(mode="json", exclude_defaults=True)
            return Response(200, json=result)
        else:
            return Response(404)

    @_check(fail_on=MockGafaelfawrAction.USER_INFO)
    def _handle_user_info(
        self, request: Request, token_data: TokenData
    ) -> Response:
        if user_info := self._user_info.get(token_data.username):
            result = user_info.model_dump(mode="json", exclude_defaults=True)
            return Response(200, json=result)
        else:
            return Response(404)


async def register_mock_gafaelfawr(respx_mock: respx.Router) -> MockGafaelfawr:
    """Mock out Gafaelfawr.

    Parameters
    ----------
    respx_mock
        Mock router.

    Returns
    -------
    MockGafaelfawr
        Mock Gafaelfawr API object.
    """
    discovery_client = DiscoveryClient()
    url = await discovery_client.url_for_internal("gafaelfawr", version="v1")
    await discovery_client.aclose()
    assert url, "Service gafaelfawr (v1) not found in Repertoire"
    mock = MockGafaelfawr()
    mock.install_routes(respx_mock, url)
    return mock



"""Exceptions for the Gafaelfawr client."""

from typing import Self, override

from pydantic import ValidationError
from safir.slack.blockkit import (
    SentryEventInfo,
    SlackCodeBlock,
    SlackException,
    SlackMessage,
    SlackWebException,
)

__all__ = [
    "GafaelfawrDiscoveryError",
    "GafaelfawrError",
    "GafaelfawrNotFoundError",
    "GafaelfawrValidationError",
    "GafaelfawrWebError",
]


class GafaelfawrError(SlackException):
    """Base class for Gafaelfawr client exceptions."""


class GafaelfawrDiscoveryError(GafaelfawrError):
    """Gafaelfawr was not found in service discovery."""


class GafaelfawrValidationError(GafaelfawrError):
    """Gafaelfawr response did not validate against the expected model."""

    @classmethod
    def from_exception(cls, exc: ValidationError) -> Self:
        """Create an exception from a Pydantic parse failure.

        Parameters
        ----------
        exc
            Pydantic exception.

        Returns
        -------
        GafaelfawrValidationError
            Constructed exception.
        """
        error = f"{type(exc).__name__}: {exc!s}"
        return cls("Unable to parse reply from Gafaelfawr", error)

    def __init__(self, message: str, error: str) -> None:
        super().__init__(message)
        self._message = message
        self.error = error

    @override
    def __str__(self) -> str:
        return f"{self._message}: {self.error}"

    @override
    def to_slack(self) -> SlackMessage:
        message = super().to_slack()
        block = SlackCodeBlock(heading="Validation error", code=self.error)
        message.attachments.append(block)
        return message

    @override
    def to_sentry(self) -> SentryEventInfo:
        info = super().to_sentry()
        info.contexts["validation_info"] = {"error": self.error}
        return info


class GafaelfawrWebError(SlackWebException, GafaelfawrError):
    """An HTTP request failed."""


class GafaelfawrNotFoundError(GafaelfawrWebError):
    """An HTTP request failed with a 404 response."""



"""Tests for Gafaelfawr client models."""

from rubin.gafaelfawr import (
    GafaelfawrGroup,
    GafaelfawrNotebookQuota,
    GafaelfawrTapQuota,
    GafaelfawrUserInfo,
)


def test_memory_bytes() -> None:
    quota = GafaelfawrNotebookQuota(cpu=1.0, memory=2.0)
    assert quota.memory_bytes == 2147483648


def test_supplemental_groups() -> None:
    userinfo = GafaelfawrUserInfo(
        username="someone",
        gid=4000,
        groups=[
            GafaelfawrGroup(name="a", id=1000),
            GafaelfawrGroup(name="b", id=1001),
            GafaelfawrGroup(name="c", id=1002),
        ],
    )
    assert userinfo.supplemental_groups == [1000, 1001, 1002]


def test_to_logging_context() -> None:
    notebook = GafaelfawrNotebookQuota(cpu=1.0, memory=4.0)
    assert notebook.to_logging_context() == {"cpu": 1.0, "memory": "4.0 GiB"}
    notebook = GafaelfawrNotebookQuota(cpu=1.5, memory=6.5, spawn=False)
    assert notebook.to_logging_context() == {
        "cpu": 1.5,
        "memory": "6.5 GiB",
        "spawn": False,
    }

    tap = GafaelfawrTapQuota(concurrent=4)
    assert tap.to_logging_context() == {"concurrent": 4}



"""Constants for the Gafaelfawr client."""

from datetime import timedelta

__all__ = ["CACHE_LIFETIME", "CACHE_SIZE"]

CACHE_LIFETIME = timedelta(minutes=5)
"""How long to cache user information retrieved from Gafaelfawr.

Gafaelfawr policy says that this should not be any longer than five minutes so
that changes to the user's groups are picked up correctly.
"""

CACHE_SIZE = 1000
"""Maximum number of users whose information is cached in memory."""



"""Helper functions for reading test data."""

import json
from pathlib import Path
from typing import Any

from rubin.gafaelfawr import GafaelfawrUserInfo

__all__ = [
    "data_path",
    "read_test_json",
    "read_test_user_info",
]


def data_path(fragment: str) -> Path:
    """Construct a path to a test data file.

    Parameters
    ----------
    fragment
        Path relative to :file:`tests/data`.

    Returns
    -------
    Path
        Full path to file.
    """
    return Path(__file__).parent.parent / "data" / fragment


def read_test_json(fragment: str) -> Any:
    """Read test data as JSON and return its decoded form.

    Parameters
    ----------
    fragment
        Path relative to :file:`tests/data`.

    Returns
    -------
    typing.Any
        Parsed contents of the file.
    """
    path = data_path(fragment + ".json")
    with path.open("r") as f:
        return json.load(f)


def read_test_user_info(username: str) -> GafaelfawrUserInfo:
    """Read test user information for a user.

    Parameters
    ----------
    username
        Username whose test data should be read.

    Returns
    -------
    GafaelfawrUserInfo
        Parsed contents of the file.
    """
    raw_user_info = read_test_json(f"userinfo/{username}")
    return GafaelfawrUserInfo.model_validate(raw_user_info)



"""Tests for the Gafaelfawr client."""

from datetime import UTC, datetime, timedelta
from pathlib import Path

import pytest
import respx
from rubin.repertoire import register_mock_discovery

from rubin.gafaelfawr import (
    GafaelfawrClient,
    GafaelfawrDiscoveryError,
    GafaelfawrGroup,
    GafaelfawrNotFoundError,
    GafaelfawrUserInfo,
    GafaelfawrWebError,
    MockGafaelfawr,
    MockGafaelfawrAction,
)

from .support.data import read_test_user_info


@pytest.mark.asyncio
async def test_create_token(mock_gafaelfawr: MockGafaelfawr) -> None:
    token = mock_gafaelfawr.create_token("admin", scopes=["admin:token"])
    client = GafaelfawrClient()

    # Create a minimal token with just a username.
    service_token = await client.create_service_token(
        token, "bot-service", scopes=[]
    )
    userinfo = await client.get_user_info(service_token)
    assert userinfo == GafaelfawrUserInfo(username="bot-service")

    # This token has no scopes so should not be able to retrieve user
    # information by username.
    with pytest.raises(GafaelfawrWebError) as exc_info:
        await client.get_user_info(service_token, "bot-service")
    assert exc_info.value.status == 403

    # Create a token with all of the fields set.
    service_token = await client.create_service_token(
        token,
        "bot-service",
        scopes=["admin:userinfo"],
        expires=datetime.now(tz=UTC) + timedelta(days=1),
        name="Some bot user",
        uid=4000,
        gid=4001,
        groups=[GafaelfawrGroup(name="group", id=5000)],
    )
    expected = GafaelfawrUserInfo(
        username="bot-service",
        name="Some bot user",
        uid=4000,
        gid=4001,
        groups=[GafaelfawrGroup(name="group", id=5000)],
    )
    assert await client.get_user_info(service_token) == expected

    # This token has admin:userinfo, so can get user information by username.
    assert await client.get_user_info(service_token, "bot-service") == expected


@pytest.mark.asyncio
async def test_missing_discovery(
    monkeypatch: pytest.MonkeyPatch, respx_mock: respx.Router
) -> None:
    monkeypatch.setenv("REPERTOIRE_BASE_URL", "https://example.com/repertoire")
    path = Path(__file__).parent / "data" / "empty.json"
    register_mock_discovery(respx_mock, path)
    client = GafaelfawrClient()

    # If the Gafaelfawr service (v1) is not included in the discovery
    # information, any client call should raise GafaelfawrDiscoveryError.
    with pytest.raises(GafaelfawrDiscoveryError):
        await client.get_user_info("some-token")


@pytest.mark.asyncio
async def test_userinfo_by_username(mock_gafaelfawr: MockGafaelfawr) -> None:
    token = mock_gafaelfawr.create_token("admin", scopes=["admin:userinfo"])
    user_info = read_test_user_info("someuser")
    client = GafaelfawrClient()

    # If no data is available, should raise GafaelfawrNotFoundError.
    with pytest.raises(GafaelfawrNotFoundError):
        await client.get_user_info(token, "someuser")

    # HTTP errors should raise GafaelfawrWebError.
    mock_gafaelfawr.fail_on("someuser", MockGafaelfawrAction.USER_INFO)
    with pytest.raises(GafaelfawrWebError):
        await client.get_user_info(token, "someuser")

    # Register some user information and try again.
    mock_gafaelfawr.fail_on("someuser", [])
    mock_gafaelfawr.set_user_info("someuser", user_info)
    assert await client.get_user_info(token, "someuser") == user_info


@pytest.mark.asyncio
async def test_cache_by_username(mock_gafaelfawr: MockGafaelfawr) -> None:
    token = mock_gafaelfawr.create_token("admin", scopes=["admin:userinfo"])
    user_info = read_test_user_info("someuser")
    empty_user_info = GafaelfawrUserInfo(username="someuser")
    client = GafaelfawrClient()

    # Register the empty object at first.
    mock_gafaelfawr.set_user_info("someuser", empty_user_info)
    assert await client.get_user_info(token, "someuser") == empty_user_info

    # Changing the underlying user information shouldn't change the result
    # since it is still cached.
    mock_gafaelfawr.set_user_info("someuser", user_info)
    assert await client.get_user_info(token, "someuser") == empty_user_info

    # Clearing the cache should result in the new data.
    await client.clear_cache()
    assert await client.get_user_info(token, "someuser") == user_info


@pytest.mark.asyncio
async def test_userinfo_by_token(mock_gafaelfawr: MockGafaelfawr) -> None:
    token = mock_gafaelfawr.create_token("someuser")
    user_info = read_test_user_info("someuser")
    client = GafaelfawrClient()

    # If no data is available, should raise GafaelfawrNotFoundError.
    with pytest.raises(GafaelfawrNotFoundError):
        await client.get_user_info(token)

    # HTTP errors should raise GafaelfawrWebError.
    mock_gafaelfawr.fail_on("someuser", MockGafaelfawrAction.USER_INFO)
    with pytest.raises(GafaelfawrWebError):
        await client.get_user_info(token)

    # Register some user information and try again.
    mock_gafaelfawr.fail_on("someuser", [])
    mock_gafaelfawr.set_user_info("someuser", user_info)
    assert await client.get_user_info(token) == user_info


@pytest.mark.asyncio
async def test_cache_by_token(mock_gafaelfawr: MockGafaelfawr) -> None:
    token = mock_gafaelfawr.create_token("someuser")
    user_info = read_test_user_info("someuser")
    empty_user_info = GafaelfawrUserInfo(username="someuser")
    client = GafaelfawrClient()

    # Register the empty object at first.
    mock_gafaelfawr.set_user_info("someuser", empty_user_info)
    assert await client.get_user_info(token) == empty_user_info

    # Changing the underlying user information shouldn't change the result
    # since it is still cached.
    mock_gafaelfawr.set_user_info("someuser", user_info)
    assert await client.get_user_info(token) == empty_user_info

    # Clearing the cache should result in the new data.
    await client.clear_cache()
    assert await client.get_user_info(token) == user_info



"""Tests for Gafaelfawr client FastAPI dependencies."""

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Annotated

import pytest
from asgi_lifespan import LifespanManager
from fastapi import Depends, FastAPI, Header
from httpx import ASGITransport, AsyncClient
from safir.dependencies.http_client import http_client_dependency

from rubin.gafaelfawr import (
    GafaelfawrClient,
    GafaelfawrUserInfo,
    MockGafaelfawr,
    gafaelfawr_dependency,
)


@pytest.mark.asyncio
async def test_dependency(mock_gafaelfawr: MockGafaelfawr) -> None:
    token = mock_gafaelfawr.create_token("user")
    mock_gafaelfawr.set_user_info("user", GafaelfawrUserInfo(username="user"))
    cached_client = None

    @asynccontextmanager
    async def lifespan(app: FastAPI) -> AsyncGenerator[None]:
        yield
        await http_client_dependency.aclose()

    app = FastAPI(lifespan=lifespan)

    @app.get("/")
    async def get_root(
        x_auth_request_token: Annotated[str, Header()],
        gafaelfawr: Annotated[
            GafaelfawrClient, Depends(gafaelfawr_dependency)
        ],
    ) -> None:
        nonlocal cached_client
        if cached_client is None:
            cached_client = gafaelfawr
        assert gafaelfawr == cached_client
        user_info = await gafaelfawr.get_user_info(x_auth_request_token)
        assert user_info == GafaelfawrUserInfo(username="user")

    async with LifespanManager(app):
        async with AsyncClient(
            base_url="https://example.com/", transport=ASGITransport(app=app)
        ) as client:
            r = await client.get("/", headers={"X-Auth-Request-Token": token})
            assert r.status_code == 200
            r = await client.get("/", headers={"X-Auth-Request-Token": token})
            assert r.status_code == 200

    # When the HTTPX client dependency is shut down and recreated, this should
    # result in a new Gafaelfawr client. Otherwise, the Gafaelfawr client
    # would try to use the closed HTTPX client.
    old_client = cached_client
    cached_client = None
    async with LifespanManager(app):
        async with AsyncClient(
            base_url="https://example.com/", transport=ASGITransport(app=app)
        ) as client:
            r = await client.get("/", headers={"X-Auth-Request-Token": token})
            assert r.status_code == 200
            assert cached_client != old_client



"""nox configuration for the Gafaelfawr client.

This subdirectory has its own noxfile because it supports a wider range of
Python versions and is matrix-tested on those versions in GitHub Actions
independent from the top-level directory. The top-level noxfile recurses into
this one as needed.
"""

import nox
from nox_uv import session

# Default sessions.
nox.options.sessions = ["typing", "test-coverage", "coverage-report"]

# Other nox defaults.
nox.options.default_venv_backend = "uv"
nox.options.reuse_existing_virtualenvs = True


@session(name="coverage-report", uv_groups=["dev"])
def coverage_report(session: nox.Session) -> None:
    """Generate a code coverage report from the test suite."""
    session.run("coverage", "report", *session.posargs)


@session(uv_groups=["dev"])
def test(session: nox.Session) -> None:
    """Test both the server and the client."""
    session.run("pytest", *session.posargs)


@session(name="test-coverage", uv_groups=["dev"])
def test_coverage(session: nox.Session) -> None:
    """Test both the server and the client."""
    session.run(
        "pytest",
        "--cov=rubin.gafaelfawr",
        "--cov-branch",
        "--cov-report=",
        *session.posargs,
    )


@session(uv_groups=["dev", "typing"])
def typing(session: nox.Session) -> None:
    """Run mypy."""
    session.run(
        "mypy",
        *session.posargs,
        "--namespace-packages",
        "--explicit-package-bases",
        "noxfile.py",
        "src",
        "tests",
        env={"MYPYPATH": "src"},
    )



"""Models for the Gafaelawr API understood by the client.

These models are intentionally not shared with the server implementation since
they may have to handle multiple versions of the server. They are simplified
compared to the server models, and only the ones starting with ``Gafaelfawr``
are exposed to users of the module.
"""

from datetime import datetime
from enum import Enum
from typing import Annotated, Any

from pydantic import BaseModel, Field

__all__ = [
    "AdminTokenRequest",
    "GafaelfawrGroup",
    "GafaelfawrNotebookQuota",
    "GafaelfawrQuota",
    "GafaelfawrTapQuota",
    "GafaelfawrUserInfo",
    "NewToken",
    "TokenData",
    "TokenType",
]


class GafaelfawrGroup(BaseModel):
    """Information about a single group."""

    name: Annotated[str, Field(title="Name of the group")]

    id: Annotated[int, Field(title="Numeric GID of the group")]


class GafaelfawrNotebookQuota(BaseModel):
    """Notebook Aspect quota information for a user."""

    cpu: Annotated[float, Field(title="CPU equivalents", examples=[4.0])]

    memory: Annotated[float, Field(title="Maximum memory use (GiB)")]

    spawn: Annotated[
        bool,
        Field(title="Spawning allowed"),
    ] = True

    @property
    def memory_bytes(self) -> int:
        """Maximum memory use in bytes."""
        return int(self.memory * 1024 * 1024 * 1024)

    def to_logging_context(self) -> dict[str, Any]:
        """Convert to variables for a structlog logging context."""
        result = {"cpu": self.cpu, "memory": f"{self.memory} GiB"}
        if not self.spawn:
            result["spawn"] = False
        return result


class GafaelfawrTapQuota(BaseModel):
    """TAP quota information for a user."""

    concurrent: Annotated[int, Field(title="Concurrent queries")]

    def to_logging_context(self) -> dict[str, Any]:
        """Convert to variables for a structlog logging context."""
        return {"concurrent": self.concurrent}


class GafaelfawrQuota(BaseModel):
    """Quota information for a user."""

    api: Annotated[
        dict[str, int],
        Field(
            title="API quotas",
            description=("Mapping of service names to requests per minute"),
        ),
    ] = {}

    notebook: Annotated[
        GafaelfawrNotebookQuota | None, Field(title="Notebook Aspect quotas")
    ] = None

    tap: Annotated[
        dict[str, GafaelfawrTapQuota], Field(title="TAP quotas")
    ] = {}


class GafaelfawrUserInfo(BaseModel):
    """Information about a user."""

    username: Annotated[str, Field(..., title="Username")]

    name: Annotated[str | None, Field(title="Preferred full name")] = None

    email: Annotated[str | None, Field(title="Email address")] = None

    uid: Annotated[int | None, Field(title="UID number")] = None

    gid: Annotated[int | None, Field(title="Primary GID")] = None

    groups: Annotated[list[GafaelfawrGroup], Field(title="Groups")] = []

    quota: Annotated[GafaelfawrQuota | None, Field(title="Quota")] = None

    @property
    def supplemental_groups(self) -> list[int]:
        """Supplemental GIDs."""
        return [g.id for g in self.groups]


class TokenData(BaseModel):
    """Metadata about a token, used internally by the mock."""

    token: Annotated[str, Field(title="Associated token")]

    username: Annotated[str, Field(title="Username")]

    scopes: Annotated[set[str], Field(title="Token scopes")] = set()

    expires: Annotated[datetime | None, Field(title="Expiration time")] = None


class TokenType(Enum):
    """The class of token.

    This includes only the subset of token types used by the client.
    """

    service = "service"


class AdminTokenRequest(BaseModel):
    """A request to create a new token via the admin interface."""

    username: Annotated[str, Field(title="User for which to issue a token")]

    token_type: Annotated[TokenType, Field(title="Token type")]

    scopes: Annotated[list[str], Field(title="Token scopes")] = []

    expires: Annotated[datetime | None, Field(title="Token expiration")] = None

    name: Annotated[str | None, Field(title="Preferred full name")] = None

    email: Annotated[str | None, Field(title="Email address")] = None

    uid: Annotated[int | None, Field(title="UID number")] = None

    gid: Annotated[int | None, Field(title="Primary GID")] = None

    groups: Annotated[
        list[GafaelfawrGroup] | None,
        Field(title="Groups"),
    ] = None


class NewToken(BaseModel):
    """Response to a token creation request."""

    token: Annotated[str, Field(title="Newly-created token")]



"""Client for the Gafaelfawr authorization and identity service."""

import asyncio
from datetime import datetime, timedelta

import structlog
from cachetools import TTLCache
from httpx import AsyncClient, HTTPError, HTTPStatusError, Timeout
from pydantic import BaseModel, ValidationError
from rubin.repertoire import DiscoveryClient
from structlog.stdlib import BoundLogger

from ._constants import CACHE_LIFETIME, CACHE_SIZE
from ._exceptions import (
    GafaelfawrDiscoveryError,
    GafaelfawrNotFoundError,
    GafaelfawrValidationError,
    GafaelfawrWebError,
)
from ._models import (
    AdminTokenRequest,
    GafaelfawrGroup,
    GafaelfawrUserInfo,
    NewToken,
    TokenType,
)

__all__ = ["GafaelfawrClient"]


class GafaelfawrClient:
    """Client for the Gafaelfawr service API.

    Parameters
    ----------
    http_client
        Existing ``httpx.AsyncClient`` to use instead of creating a new one.
        This allows the caller to reuse an existing client and connection
        pool.
    discovery_client
        If given, Repertoire_ discovery client to use. Otherwise, a new client
        will be created.
    logger
        Logger to use. If not given, the default structlog logger will be
        used.
    timeout
        Timeout for Gafaelfawr operations. If not given, defaults to the
        timeout of the underlying HTTPX_ client.
    userinfo_cache_lifetime
        How long to cache user information for a token.
    userinfo_cache_size
        How many cache entries for the cache of user information by token.
    """

    def __init__(
        self,
        http_client: AsyncClient | None = None,
        *,
        discovery_client: DiscoveryClient | None = None,
        logger: BoundLogger | None = None,
        timeout: timedelta | None = None,
        userinfo_cache_lifetime: timedelta = CACHE_LIFETIME,
        userinfo_cache_size: int = CACHE_SIZE,
    ) -> None:
        self._client = http_client or AsyncClient()
        self._discovery = discovery_client or DiscoveryClient(self._client)
        self._logger = logger or structlog.get_logger()
        self._userinfo_cache_lifetime = userinfo_cache_lifetime
        self._userinfo_cache_size = userinfo_cache_size

        # Whether the HTTP client needs to be explicitly closed because we
        # created it.
        self._close_client = http_client is not None

        # The default timeout is the underlying timeout of the HTTPX client.
        if timeout is not None:
            self._timeout: float | Timeout = timeout.total_seconds()
        else:
            self._timeout = self._client.timeout

        # Maintain two caches of user information, one indexed by token (when
        # requesting user information for the user corresponding to the token)
        # and one indexed by username (when requesting user information with a
        # privileged token). Most users of the client will only make one type
        # of call, but hopefully the overhead is tiny.
        self._userinfo_token_cache: TTLCache[str, GafaelfawrUserInfo]
        self._userinfo_token_cache = TTLCache(
            userinfo_cache_size, userinfo_cache_lifetime.total_seconds()
        )
        self._userinfo_token_lock = asyncio.Lock()
        self._userinfo_username_cache: TTLCache[str, GafaelfawrUserInfo]
        self._userinfo_username_cache = TTLCache(
            userinfo_cache_size, userinfo_cache_lifetime.total_seconds()
        )
        self._userinfo_username_lock = asyncio.Lock()

    async def aclose(self) -> None:
        """Close the HTTP connection pool, if one wasn't provided.

        Only closes the pool if a new one was created. Does nothing if an
        external HTTP connection pool was passed into the constructor. The
        object must not be used after calling this method.
        """
        if self._close_client:
            await self._client.aclose()

    async def clear_cache(self) -> None:
        """Clear all internal caches."""
        async with self._userinfo_token_lock:
            self._userinfo_token_cache = TTLCache(
                self._userinfo_cache_size,
                self._userinfo_cache_lifetime.total_seconds(),
            )
        async with self._userinfo_username_lock:
            self._userinfo_username_cache = TTLCache(
                self._userinfo_cache_size,
                self._userinfo_cache_lifetime.total_seconds(),
            )

    async def create_service_token(
        self,
        token: str,
        username: str,
        *,
        scopes: list[str],
        expires: datetime | None = None,
        name: str | None = None,
        uid: int | None = None,
        gid: int | None = None,
        groups: list[GafaelfawrGroup] | None = None,
    ) -> str:
        """Create a new service token.

        Parameters
        ----------
        token
            Token to use to authenticate to the Gafaelfawr API. This token
            must have the ``admin:token`` scope.
        username
            Username for which to create a token. Must begin with ``bot-``.
        scopes
            List of scopes to grant to the new token.
        expires
            Expiration date of te new token, or `None` to create a token that
            never expires.
        name
            Full name override. If `None`, the full name will be determined
            from LDAP if configured, and otherwise not set.
        uid
            UID override. If `None`, the UID will be determined from Firestore
            or LDAP if configured, and otherwise not set.
        gid
            Primary GID override. If `None`, the primary GID will be
            determined from Firestore or LDAP if configured, and otherwise not
            set.
        groups
            Group membership override. If `None`, the group membership will be
            determined from LDAP if configured, and otherwise not set.

        Returns
        -------
        str
            Newly-created token.

        Raises
        ------
        GafaelfawrValidationError
            Raised if the response from Gafaelfawr is invalid.
        GafaelfawrWebError
            Raised if there is some problem talking to the Gafaelfawr API,
            such as an invalid token or network or service failure.
        rubin.repertoire.RepertoireError
            Raised if there was an error talking to service discovery.
        """
        url = await self._url_for("tokens")
        request = AdminTokenRequest(
            username=username,
            token_type=TokenType.service,
            scopes=scopes,
            expires=expires,
            name=name,
            uid=uid,
            gid=gid,
            groups=groups,
        )
        result = await self._post(url, NewToken, token, body=request)
        return result.token

    async def get_user_info(
        self, token: str, username: str | None = None
    ) -> GafaelfawrUserInfo:
        """Get information about the user, with caching.

        Parameters
        ----------
        token
            Token to use to authenticate to the Gafaelfawr API.
        username
            Username for which to request user information, if given. If not
            given, the user information for the user corresponding to the
            authentication token will be retrieved. This parameter may only be
            provided when authenticating with a token with ``admin:userinfo``
            scope.

        Returns
        -------
        GafaelfawrUserInfo
            User information for the user.

        Raises
        ------
        GafaelfawrNotFoundError
            Raised if no user information for the requested user could be
            found. This will always be the case when the ``username``
            parameter is provided and Gafaelfawr is not configured to use LDAP
            for user information, since in that case user information can only
            be retrieved with a user's token.
        GafaelfawrValidationError
            Raised if the response from Gafaelfawr is invalid.
        GafaelfawrWebError
            Raised if there is some problem talking to the Gafaelfawr API,
            such as an invalid token or network or service failure.
        rubin.repertoire.RepertoireError
            Raised if there was an error talking to service discovery.
        """
        if username is not None:
            if userinfo := self._userinfo_username_cache.get(username):
                return userinfo
            async with self._userinfo_username_lock:
                if userinfo := self._userinfo_username_cache.get(username):
                    return userinfo
                url = await self._url_for(f"users/{username}")
                userinfo = await self._get(url, GafaelfawrUserInfo, token)
                self._userinfo_username_cache[username] = userinfo
                return userinfo
        else:
            if userinfo := self._userinfo_token_cache.get(token):
                return userinfo
            async with self._userinfo_token_lock:
                if userinfo := self._userinfo_token_cache.get(token):
                    return userinfo
                url = await self._url_for("user-info")
                userinfo = await self._get(url, GafaelfawrUserInfo, token)
                self._userinfo_token_cache[token] = userinfo
                return userinfo

    async def _get[T: BaseModel](
        self, url: str, model: type[T], token: str
    ) -> T:
        """Make an HTTP GET request and validate the results.

        Parameters
        ----------
        url
            URL at which to make the request.
        model
            Expected type of the response.
        token
            Gafaelfawr token used for authentication.

        Returns
        -------
        pydantic.BaseModel
            Validated model of the requested type.

        Raises
        ------
        GafaelfawrNotFoundError
            Raised if the requested URL returned a 404 response.
        GafaelfawrValidationError
            Raised if the response from Gafaelfawr is invalid.
        GafaelfawrWebError
            Raised if there is some problem talking to the Gafaelfawr API,
            such as an invalid token or network or service failure.
        """
        headers = {"Authorization": f"Bearer {token}"}
        try:
            r = await self._client.get(
                url, headers=headers, timeout=self._timeout
            )
            r.raise_for_status()
            return model.model_validate(r.json())
        except HTTPError as e:
            if isinstance(e, HTTPStatusError):
                if e.response.status_code == 404:
                    raise GafaelfawrNotFoundError.from_exception(e) from e
            raise GafaelfawrWebError.from_exception(e) from e
        except ValidationError as e:
            raise GafaelfawrValidationError.from_exception(e) from e

    async def _post[T: BaseModel](
        self, url: str, model: type[T], token: str, *, body: BaseModel
    ) -> T:
        """Make an HTTP GET request and validate the results.

        Parameters
        ----------
        url
            URL at which to make the request.
        model
            Expected type of the response.
        token
            Gafaelfawr token used for authentication.
        body
            Pydantic model to send as the POST body.

        Returns
        -------
        pydantic.BaseModel
            Validated model of the requested type.

        Raises
        ------
        GafaelfawrValidationError
            Raised if the response from Gafaelfawr is invalid.
        GafaelfawrWebError
            Raised if there is some problem talking to the Gafaelfawr API,
            such as an invalid token or network or service failure.
        """
        headers = {"Authorization": f"Bearer {token}"}
        try:
            r = await self._client.post(
                url,
                headers=headers,
                json=body.model_dump(mode="json"),
                timeout=self._timeout,
            )
            r.raise_for_status()
            return model.model_validate(r.json())
        except HTTPError as e:
            raise GafaelfawrWebError.from_exception(e) from e
        except ValidationError as e:
            raise GafaelfawrValidationError.from_exception(e) from e

    async def _url_for(self, route: str) -> str:
        """Construct the URL for a Gafaelfawr API route.

        Parameters
        ----------
        route
            Route relative to the Gafaelfawr API base URL. Must not start with
            ``/``.

        Returns
        -------
        str
            Full URL to use.

        Raises
        ------
        GafaelfawrDiscoveryError
            Raised if Gafaelfawr is missing from service discovery.
        rubin.repertoire.RepertoireError
            Raised if there was an error talking to service discovery.
        """
        base_url = await self._discovery.url_for_internal(
            "gafaelfawr", version="v1"
        )
        if not base_url:
            msg = "gafaelfawr (v1) service not found in service discovery"
            raise GafaelfawrDiscoveryError(msg)
        return f"{base_url.rstrip('/')}/{route}"
